import logging
from contextlib import contextmanager
from multiprocessing import Lock as LockMp
from multiprocessing import Value
from threading import Lock


class RwLock:
    def __init__(self, name=None, write_first=True, debug=False,
                 cross_process=True):
        self.name = name if name else self.__class__.__name__
        self.log = logging.getLogger(self.name)
        self.log.propagate = debug
        self._cross_process = cross_process

        # prevent writer starvation. block new read when someone want to write.
        self._write_first = write_first

        if self._cross_process:
            self._write_lock = LockMp()
            if self._write_first:
                self._read_lock = LockMp()
            self._read_cnt = Value('i', 0)
        else:
            self._write_lock = Lock()
            if self._write_first:
                self._read_lock = Lock()
            self._read_cnt_lock = Lock()
            self._read_cnt = 0

    def acquire_r(self):
        self.log.debug("")
        if self._write_first:
            self.log.debug("wait writter")
            self._read_lock.acquire()
            self.log.debug("wait writter done")
            self._read_lock.release()

        if self._cross_process:
            with self._read_cnt.get_lock():
                self._read_cnt.value += 1
                self.log.debug("read count+ = %s", self._read_cnt.value)
                if self._read_cnt.value == 1:
                    self._write_lock.acquire()
        else:
            with self._read_cnt_lock:
                self._read_cnt += 1
                self.log.debug("read count+ = %s", self._read_cnt)
                if self._read_cnt == 1:
                    self._write_lock.acquire()

    def release_r(self):
        self.log.debug("")
        if self._cross_process:
            with self._read_cnt.get_lock():
                self._read_cnt.value -= 1
                self.log.debug("read count- = %s", self._read_cnt.value)
                if self._read_cnt.value == 0:
                    self._write_lock.release()
        else:
            with self._read_cnt_lock:
                self._read_cnt -= 1
                self.log.debug("read count- = %s", self._read_cnt)
                if self._read_cnt == 0:
                    self._write_lock.release()

    def acquire_w(self):
        self.log.debug("")
        if self._write_first:
            self.log.debug("block new read")
            self._read_lock.acquire()
        self._write_lock.acquire()

    def release_w(self):
        self.log.debug("")
        self._write_lock.release()
        if self._write_first:
            self._read_lock.release()
            self.log.debug("un-block new read")

    @contextmanager
    def lock_r(self):
        try:
            self.acquire_r()
            yield
        finally:
            self.release_r()

    @contextmanager
    def lock_w(self):
        try:
            self.acquire_w()
            yield
        finally:
            self.release_w()



import logging
from time import sleep
from unittest import TestCase
from rwlock import RwLock
from time import time as now


class TestRwLock(TestCase):
    TIME_ACCURACY = 0.2
    TIME_TICK = 0.1

    def setUp(self):
        if not hasattr(self, 'log'):
            self.log = logging.getLogger(self.__class__.__name__)

    def tearDown(self):
        pass

    def chk_time_match(self, initial_time, expected_start_time):
        t = now() - initial_time
        diff = abs(t - expected_start_time)

        self.log.info("t=%s, diff=%s", t, diff)
        self.assertTrue((diff / expected_start_time) < self.TIME_ACCURACY)

    def reader(self, name, lock, initial_time, start_time, running_time,
               expected_start_time):
        start_time *= self.TIME_TICK
        running_time *= self.TIME_TICK
        expected_start_time *= self.TIME_TICK

        diff = (initial_time + start_time) - now()
        sleep(0 if diff < 0 else diff)

        self.log.debug("%s: in", name)
        with lock.lock_r():
            self.log.debug("%s: start", name)
            self.chk_time_match(initial_time, expected_start_time)
            sleep(running_time)
            self.log.debug("%s: end", name)

    def writer(self, name, lock, initial_time, start_time, running_time,
               expected_start_time):
        start_time *= self.TIME_TICK
        running_time *= self.TIME_TICK
        expected_start_time *= self.TIME_TICK

        diff = (initial_time + start_time) - now()
        sleep(0 if diff < 0 else diff)

        self.log.debug("%s: in", name)
        with lock.lock_w():
            self.log.debug("%s: start", name)
            self.chk_time_match(initial_time, expected_start_time)
            sleep(running_time)
            self.log.debug("%s: end", name)

    def _test_rwlock_write_first(self, cross_process):
        if cross_process:
            from multiprocessing import Process as Runner
        else:
            from threading import Thread as Runner

        rwlock = RwLock(write_first=True, debug=True,
                        cross_process=cross_process)

        runner = []
        initial_time = now()
        runner.append(Runner(target=self.reader,
                             args=('r1', rwlock, initial_time, 1, 2, 1)))
        runner.append(Runner(target=self.reader,
                             args=('r2', rwlock, initial_time, 2, 4, 2)))
        runner.append(Runner(target=self.reader,
                             args=('r3', rwlock, initial_time, 5, 2, 7)))
        runner.append(Runner(target=self.writer,
                             args=('w1', rwlock, initial_time, 4, 1, 6)))
        [w.start() for w in runner]
        [w.join() for w in runner]


    def _test_rwlock(self, cross_process):
        if cross_process:
            from multiprocessing import Process as Runner
        else:
            from threading import Thread as Runner

        self.log.info("cross_process=%s", cross_process)

        rwlock = RwLock(write_first=False, debug=True,
                        cross_process=cross_process)
        runner = []
        initial_time = now()
        runner.append(Runner(target=self.reader,
                             args=('r1', rwlock, initial_time, 1, 2, 1)))
        runner.append(Runner(target=self.reader,
                             args=('r2', rwlock, initial_time, 2, 4, 2)))
        runner.append(Runner(target=self.reader,
                             args=('r3', rwlock, initial_time, 5, 2, 5)))
        runner.append(Runner(target=self.writer,
                             args=('w1', rwlock, initial_time, 4, 1, 7)))
        [w.start() for w in runner]
        [w.join() for w in runner]

    def test_rwlock_process(self):
        self._test_rwlock(cross_process=True)

    def test_rwlock_thread(self):
        self._test_rwlock(cross_process=False)

    def test_rwlock_write_first_process(self):
        self._test_rwlock_write_first(cross_process=True)

    def test_rwlock_write_first_thread(self):
        self._test_rwlock_write_first(cross_process=False)




import re
import os
import subprocess

from setuptools import setup, find_packages

##########################################################
DEFAULT_VERSION = "0.0.0"
AUTHOR = 'ChihYing_Lin'
EMAIL = ''
PACKAGE_NAME = 'cy_rwlock'
URL = 'https://github.com/cy-arduino/' + PACKAGE_NAME
LICENSE = 'LGPL'
DESCRIPTION = 'RwLock: Reader-Writer lock'
LONG_DESCRIPTION_FILE = 'README.md'
LONG_DESCRIPTION_TYPE = 'text/markdown'
KEYWORDS = 'rwlock'
INSTALL_REQUIRES = []
CLASSIFIERS = [
        'Natural Language :: English',
        'Programming Language :: Python :: 2.7',
        'Programming Language :: Python :: 3.5',
        'Programming Language :: Python :: 3.6',
        'Programming Language :: Python :: 3.7',
        'Programming Language :: Python :: 3.8',
        'Topic :: Software Development :: Libraries',
    ]
##########################################################


# convert version from git tag to pypi style
# V0.1-3-g908f162 -> V0.1.post3
def convert_version(git_version):
    print("ori version: {}".format(git_version))

    pattern = re.compile(
        r"^[vV]*(?P<main>[0-9.]+)(-(?P<post>[0-9]+))?(-.+)?$")
    ver = pattern.search(git_version)

    if not ver:
        print("invalid version! return default version {}".format(
            DEFAULT_VERSION))
        return DEFAULT_VERSION

    new_ver = ver.group('main')

    if ver.group('post'):
        new_ver += ".{}".format(ver.group('post'))
    else:
        new_ver += ".0"

    print("new version: {}".format(new_ver))
    return new_ver


def get_pypi_version():
    try:
        version = subprocess.check_output(
            'git describe --tags', shell=True).rstrip().decode('utf-8')
        version = convert_version(version)
    except subprocess.CalledProcessError:
        version = DEFAULT_VERSION

    return version


def read_file(file_name):
    # noinspection PyBroadException
    try:
        cur_path = os.path.abspath(os.path.dirname(__file__))
        with open(os.path.join(cur_path, file_name)) as f:
            long_description = f.read()
    except Exception:
        long_description = ""
    return long_description


setup(name=PACKAGE_NAME,
      version=get_pypi_version(),
      description=DESCRIPTION,
      url=URL,
      author=AUTHOR,
      author_email=EMAIL,
      license=LICENSE,
      packages=find_packages(exclude=['tests', 'test_*']),
      long_description=read_file(LONG_DESCRIPTION_FILE),
      long_description_content_type=LONG_DESCRIPTION_TYPE,
      install_requires=INSTALL_REQUIRES,
      classifiers=CLASSIFIERS,
      keywords=KEYWORDS,
      zip_safe=False)
